from random import randint
from typing import NamedTuple


class RollInfo(NamedTuple):
    number_of_dice: int
    dice_sides: int
    bonus: float = 0


class Roll:
    def __init__(self, number_of_dice: int, dice_sides: int, bonus: float) -> None:
        self.number_of_dice = number_of_dice
        self.dice_sides = dice_sides
        self.bonus = bonus
        self._individual_results: tuple[int, ...] = ()

        self._roll()

    @classmethod
    def from_roll_info(cls, roll_info: RollInfo) -> "Roll":
        return cls(
            number_of_dice=roll_info.number_of_dice,
            dice_sides=roll_info.dice_sides,
            bonus=roll_info.bonus,
        )

    def _roll(self):
        self._individual_results = tuple(
            (randint(1, self.dice_sides) for _ in range(self.number_of_dice))
        )

    @property
    def result(self) -> float:
        return sum(self.individual_results) + self.bonus

    @property
    def individual_results(self) -> tuple[float, ...]:
        return self._individual_results



from .format import roll_as_text, roll_as_markdown_text
from .parse import parse_input, validate_roll_info
from .roll import Roll


def roll_string_from_input(
    input_string: str,
    number_of_dice_limit: int = 0,
    dice_sides_limit: int = 0,
    bonus_absolute_value_limit: int = 0,
) -> str:
    roll_info = parse_input(input_string)
    validate_roll_info(
        roll_info=roll_info,
        number_of_dice_limit=number_of_dice_limit,
        dice_sides_limit=dice_sides_limit,
        bonus_absolute_value_limit=bonus_absolute_value_limit,
    )
    roll = Roll.from_roll_info(roll_info)
    return roll_as_text(roll)


def markdown_roll_string_from_input(
    input_string: str,
    number_of_dice_limit: int = 0,
    dice_sides_limit: int = 0,
    bonus_absolute_value_limit: int = 0,
) -> str:
    roll_info = parse_input(input_string)
    validate_roll_info(
        roll_info=roll_info,
        number_of_dice_limit=number_of_dice_limit,
        dice_sides_limit=dice_sides_limit,
        bonus_absolute_value_limit=bonus_absolute_value_limit,
    )
    roll = Roll.from_roll_info(roll_info)
    return roll_as_markdown_text(roll)



from typing import Union

from .roll import Roll


def roll_as_text(roll: Roll) -> str:
    bonus = getattr(roll, "bonus", 0)
    request_bonus, after_rolls_bonus = _bonus_strings_from_bonus(bonus)

    individual_results_text = " ".join((str(r) for r in roll.individual_results))

    return (
        f"{roll.number_of_dice}d{roll.dice_sides}{request_bonus}, "
        f"tulos: [{individual_results_text}]{after_rolls_bonus} "
        f"= {as_float_if_has_decimals(roll.result)}"
    )


def roll_as_markdown_text(roll: Roll) -> str:
    bonus = getattr(roll, "bonus", 0)
    request_bonus, after_rolls_bonus = _bonus_strings_from_bonus(bonus)

    individual_results_text = " ".join((str(r) for r in roll.individual_results))

    return (
        f"`{roll.number_of_dice}d{roll.dice_sides}{request_bonus}`, "
        f"tulos: `[{individual_results_text}]{after_rolls_bonus} "
        f"= {as_float_if_has_decimals(roll.result)}`"
    )


def _bonus_strings_from_bonus(bonus: float) -> tuple[str, str]:
    """Return:
    (request bonus, after rolls bonus)
    Request bonus: number with sign
                   e.g. "+3", "-1"
                   or empty string if zero
    After rolls bonus: number with sign and spaces before and after the sign
                       e.g. " + 3", " - 1"
                       or empty string if zero
    Numbers should be formatted as integers unless they have decimals
    In that case they should be formatted as floats"""
    if bonus:
        bonus = as_float_if_has_decimals(bonus)
        request_bonus = number_as_string_with_sign(bonus)
        after_rolls_bonus = number_as_string_with_spaces_around_sign(bonus)
    else:
        request_bonus = ""
        after_rolls_bonus = ""
    return request_bonus, after_rolls_bonus


def as_float_if_has_decimals(number: Union[int, float]) -> Union[int, float]:
    return int(number) if int(number) == number else number


def number_as_string_with_spaces_around_sign(number: float) -> str:
    """Return e.g. " + 3", " - 2" """
    number_as_string = number_as_string_with_sign(number)
    for sign in ("+", "-"):
        number_as_string = number_as_string.replace(sign, f" {sign} ")
    return number_as_string


def number_as_string_with_sign(number: float) -> str:
    """Return e.g. "+3", "-2" """
    return f"{number:+}"



import re

from typing import Literal

from .roll import RollInfo


def parse_input(input_string: str) -> RollInfo:
    input_string = input_string.lower()
    # Strip all whitespace
    input_string = re.sub(r"\s+", "", input_string)

    try:
        dice, bonus = _parse_bonus(input_string)
    except ValueError:
        raise ValueError("Virhe syötteen käsittelyssä")

    try:
        number_of_dice_text, dice_sides_text = dice.split("d")
    except ValueError:
        raise ValueError("Virhe syötteen käsittelyssä")

    try:
        number_of_dice = int(number_of_dice_text)
    except ValueError:
        raise ValueError("Virhe syötteen käsittelyssä")

    try:
        dice_sides = int(dice_sides_text)
    except ValueError:
        raise ValueError("Virhe syötteen käsittelyssä")

    if dice_sides < 1:
        raise ValueError("Virhe syötteen käsittelyssä")

    return RollInfo(
        number_of_dice=number_of_dice,
        dice_sides=dice_sides,
        bonus=bonus,
    )


def validate_roll_info(
    roll_info: RollInfo,
    number_of_dice_limit: int = 0,
    dice_sides_limit: int = 0,
    bonus_absolute_value_limit: int = 0,
) -> None:
    """Raise ValueErrors if values are higher than the limit"""
    if number_of_dice_limit and roll_info.number_of_dice > number_of_dice_limit:
        raise ValueError("Liian monta noppaa")
    if dice_sides_limit and roll_info.dice_sides > dice_sides_limit:
        raise ValueError("Nopilla liian monta sivua")
    if bonus_absolute_value_limit:
        if roll_info.bonus > bonus_absolute_value_limit:
            raise ValueError("Bonus liian suuri")
        if roll_info.bonus < -bonus_absolute_value_limit:
            raise ValueError("Bonus liian pieni")


def _parse_bonus(input_string: str) -> tuple[str, float]:
    """Return (dice part, bonus)
    e.g. ("5d6", 2), ("3d10", -3)"""
    sign = _sign_in_string(input_string)
    if not sign:
        return input_string, 0

    split = re.split("[+-]", input_string)
    return split[0], float(f"{sign}{split[1]}")


def _sign_in_string(input_string: str) -> Literal["+", "-", ""]:
    if "+" in input_string:
        return "+"
    if "-" in input_string:
        return "-"
    return ""



#!/usr/bin/env python
import os
import shutil
import sys
from codecs import open

from setuptools import setup

CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (3, 8)

if CURRENT_PYTHON < REQUIRED_PYTHON:
    sys.stderr.write(
        f"Unsupported Python version - requires at least "
        f"{REQUIRED_PYTHON[0]}.{REQUIRED_PYTHON[1]}"
    )
    sys.exit(1)

# 'setup.py publish' shortcut.
if sys.argv[-1] == "publish":
    shutil.rmtree("build")
    os.system("python setup.py sdist bdist_wheel")
    os.system("twine upload dist/*")
    sys.exit()

requires: list[str] = []

here = os.path.abspath(os.path.dirname(__file__))

with open("README.md", "r", "utf-8") as f:
    readme = f.read()

setup(
    name="konso_dice_roller",
    version="0.0.4",
    description="A simple dice roller library",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Tomi Javanainen",
    author_email="tomi.javanainen@gmail.com",
    url="https://github.com/konsou/konso_dice_roller",
    packages=["konso_dice_roller"],
    package_data={"": ["LICENSE"]},
    package_dir={"konso_dice_roller": "konso_dice_roller"},
    include_package_data=True,
    python_requires=">=3.8, <4",
    install_requires=requires,
    license="MIT",
    zip_safe=False,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: Finnish",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries",
    ],
    project_urls={
        "Source": "https://github.com/konsou/konso_dice_roller",
    },
)
